/*
 * File:        TFT, keypad, DAC, LED, PORT EXPANDER test
 *              With serial interface to PuTTY console
 * Author:      Bruce Land
 * For use with Sean Carroll's Big Board
 * Target PIC:  PIC32MX250F128B
 */

////////////////////////////////////
// clock AND protoThreads configure!
// You MUST check this file!
#include "config_1_2_2.h"
// threading library
#include "pt_cornell_1_2_2.h"
// yup, the expander
#include "port_expander_brl4.h"

////////////////////////////////////
// graphics libraries
// SPI channel 1 connections to TFT
#include "tft_master.h"
#include "tft_gfx.h"
// need for rand function
#include <stdlib.h>
// need for sin function
#include <math.h>
////////////////////////////////////

// lock out timer 2 interrupt during spi comm to port expander
// This is necessary if you use the SPI2 channel in an ISR.
#define start_spi2_critical_section INTEnable(INT_T2, 0)
#define end_spi2_critical_section INTEnable(INT_T2, 1)


////////////////////////////////////

/* Demo code for interfacing TFT (ILI9340 controller) to PIC32
 * The library has been modified from a similar Adafruit library
 */
// Adafruit data:
/***************************************************
  This is an example sketch for the Adafruit 2.2" SPI display.
  This library works with the Adafruit 2.2" TFT Breakout w/SD card
  ----> http://www.adafruit.com/products/1480

  Check out the links above for our tutorials and wiring diagrams
  These displays use SPI to communicate, 4 or 5 pins are required to
  interface (RST is optional)
  Adafruit invests time and resources providing this open source code,
  please support Adafruit and open-source hardware by purchasing
  products from Adafruit!

  Written by Limor Fried/Ladyada for Adafruit Industries.
  MIT license, all text above must be included in any redistribution
 ****************************************************/

int modeFlag = 0; // 0 means direct control mode, 1 means program mode, 2 means show mode, -1 means arbitrary mode
int bufferCounter = 0; //bufferCounter to count how many buffer
char receiveBuffer[30];
int drawCounter = 0;
int msCounter = 0, msCounter2 = 0, showCounter = 0, showNum = 1, doubleCounter = 0;
int show8Counter = 0, show8Counter2 = 0;
int show9Counter = 0, show9up = 1;
int show11Counter = 0;
int showBuffer[12];
//char programArray[15][20];
void setValvesProgramMode(int baseValue);
void stopShowBuffer(void);
void setValvesShow(void);
//== Timer 2 interrupt handler ===========================================

void __ISR(_TIMER_2_VECTOR, ipl2) Timer2Handler(void) {
    mT2ClearIntFlag();
    msCounter++;
    msCounter2++;
    if (msCounter == 100) { //16ms for 50m or 0.32s of free fall per 20 interval
        if (modeFlag == 1) {
            if (drawCounter >= 0) {
                setValvesProgramMode(drawCounter * 12);
                drawCounter--;
            } else { // drawCounter == -1, finish executing
                //modeFlag = 2; //Idle mode
                drawCounter = 19; //debug
            }
            //sprintf(PT_send_buffer, "%d \n\r", drawCounter);
            //printLine2(12, PT_send_buffer, ILI9340_YELLOW , ILI9340_BLACK);
        }
        if (modeFlag == 2) {
            //Show 10; fast flashing
            if (showNum == 10) {
                if (showCounter < 50) {
                    stopShowBuffer();
                    if (doubleCounter == 0) {
                        showBuffer[0] = 0;
                        showBuffer[2] = 0;
                        showBuffer[4] = 0;
                        showBuffer[6] = 0;
                        showBuffer[8] = 0;
                        showBuffer[10] = 0;
                        doubleCounter = 1;
                    } else {
                        showBuffer[1] = 0;
                        showBuffer[3] = 0;
                        showBuffer[5] = 0;
                        showBuffer[7] = 0;
                        showBuffer[9] = 0;
                        showBuffer[11] = 0;
                        doubleCounter = 0;
                    }
                    setValvesShow();
                    showCounter++;
                } else {
                    showNum = 8;
                    showCounter = 5;
                    doubleCounter = 0;
                    stopShowBuffer();
                }
            }
            //Show 8: Falling a triangle
            if (showNum == 8) {
                if (show8Counter < 5) {
                    if (showCounter >= 0) {
                        showBuffer[showCounter] = 0;
                        showBuffer[11 - showCounter] = 0;
                        setValvesShow();
                        showCounter--;
                    } else {
                        if (show8Counter2 < 6) show8Counter2++;
                        else {
                            show8Counter++;
                            stopShowBuffer();
                            showCounter = 5;
                            show8Counter2 = 0;
                        }
                    }
                } else {
                    showNum = 9;
                    showCounter = 0;
                    doubleCounter = 0;
                }
            }//---End of show 8
            //Show 9: Falling nice effect
            if (showNum == 9) {
                if (show9Counter < 5) {
                    if (show9up) { //Rising stage
                        if (showCounter < 6) {
                            stopShowBuffer();
                            showBuffer[showCounter] = 0;
                            showBuffer[11 - showCounter] = 0;
                            setValvesShow();
                            showCounter++;
                        } else {
                            show9up = 0;
                        }
                    } else { //Falling stage
                        if (showCounter >= 0) {
                            stopShowBuffer();
                            showBuffer[showCounter] = 0;
                            showBuffer[11 - showCounter] = 0;
                            setValvesShow();
                            showCounter--;
                        } else {
                            show9up = 1;
                            show9Counter++;
                        }
                    }
                } else {
                    showNum = 11;
                    showCounter = 0;
                    doubleCounter = 0;
                }
            }//---End of show 9
            //Show 11: The End
            if (showNum == 11) {
                if (show11Counter < 50) {
                    for (showCounter = 0; showCounter < 12; showCounter++) {
                        showBuffer[showCounter] = 0;
                    }
                    setValvesShow();
                    show11Counter++;
                } else {
                    stopShowBuffer(); //Done
                    setValvesShow();
                }
            }
        }
        msCounter = 0; //reset msCounter
    }
    if (modeFlag == 2) {
        if (msCounter2 == 600) {
            //Show 1: Walking from left to right (0 to 11)
            if (showNum == 1) {
                if (showCounter < 12) {
                    stopShowBuffer();
                    showBuffer[showCounter] = 0;
                    setValvesShow();
                    showCounter++;
                } else {
                    showNum = 2;
                    showCounter = 11;
                }
            }
            //Show 2: Walking from right to left (11 to 0)
            if (showNum == 2) {
                if (showCounter >= 0) {
                    stopShowBuffer();
                    showBuffer[showCounter] = 0;
                    setValvesShow();
                    showCounter--;
                } else {
                    showNum = 3;
                    showCounter = 0;
                }
            }
            //Show 3: Walking toward the center (0 to 5 and 11 to 6)
            if (showNum == 3) {
                if (showCounter < 6) {
                    if (doubleCounter == 0) doubleCounter = 1;
                    else {
                        showBuffer[showCounter] = 0;
                        showBuffer[11 - showCounter] = 0;
                        setValvesShow();
                        showCounter++;
                        doubleCounter = 0;
                    }
                } else {
                    showNum = 4;
                    showCounter = 5;
                    doubleCounter = 0;
                }
            }
            //Show 4: Leaving from the center (5 to 0 and 6 to 11)
            if (showNum == 4) {
                if (showCounter >= 0) {
                    if (doubleCounter == 0) doubleCounter = 1;
                    else {
                        showBuffer[showCounter] = 1;
                        showBuffer[11 - showCounter] = 1;
                        setValvesShow();
                        showCounter--;
                        doubleCounter = 0;
                    }
                } else {
                    showNum = 5;
                    showCounter = 5;
                    doubleCounter = 0;
                }
            }
            //Show 5: Watering from the center to edge (5 to 0 and 6 to 11)
            if (showNum == 5) {
                if (showCounter >= 0) {
                    if (doubleCounter == 0) doubleCounter = 1;
                    else {
                        showBuffer[showCounter] = 0;
                        showBuffer[11 - showCounter] = 0;
                        setValvesShow();
                        showCounter--;
                        doubleCounter = 0;
                    }
                } else {
                    showNum = 6;
                    showCounter = 0;
                    doubleCounter = 0;
                }
            }
            //Show 6: Leaving from the edge to center (0 to 5 and 11 to 6)
            if (showNum == 6) {
                if (showCounter < 6) {
                    if (doubleCounter == 0) doubleCounter = 1;
                    else {
                        showBuffer[showCounter] = 1;
                        showBuffer[11 - showCounter] = 1;
                        setValvesShow();
                        showCounter++;
                        doubleCounter = 0;
                    }
                } else {
                    showNum = 7;
                    showCounter = 0;
                    doubleCounter = 0;
                }
            }
            //Show 7: FLash flash effect
            if (showNum == 7) {
                if (showCounter < 10) {
                    stopShowBuffer();
                    if (doubleCounter == 0) {
                        showBuffer[0] = 0;
                        showBuffer[2] = 0;
                        showBuffer[4] = 0;
                        showBuffer[6] = 0;
                        showBuffer[8] = 0;
                        showBuffer[10] = 0;
                        doubleCounter = 1;
                    } else {
                        showBuffer[1] = 0;
                        showBuffer[3] = 0;
                        showBuffer[5] = 0;
                        showBuffer[7] = 0;
                        showBuffer[9] = 0;
                        showBuffer[11] = 0;
                        doubleCounter = 0;
                    }
                    setValvesShow();
                    showCounter++;
                } else {
                    showNum = 10;
                    showCounter = 0;
                    doubleCounter = 0;
                    stopShowBuffer();
                }
            }
            msCounter2 = 0;
        }
    }
}

// === print a line on TFT =====================================================
// print a line on the TFT
// string buffer
char buffer[60];

void printLine(int line_number, char* print_buffer, short text_color, short back_color) {
    // line number 0 to 31 
    /// !!! assumes tft_setRotation(0);
    // print_buffer is the string to print
    int v_pos;
    v_pos = line_number * 10;
    // erase the pixels
    tft_fillRoundRect(0, v_pos, 239, 8, 1, back_color); // x,y,w,h,radius,color
    tft_setTextColor(text_color);
    tft_setCursor(0, v_pos);
    tft_setTextSize(1);
    tft_writeString(print_buffer);
}

void printLine2(int line_number, char* print_buffer, short text_color, short back_color) {
    // line number 0 to 31 
    /// !!! assumes tft_setRotation(0);
    // print_buffer is the string to print
    int v_pos;
    v_pos = line_number * 20;
    // erase the pixels
    tft_fillRoundRect(0, v_pos, 239, 16, 1, back_color); // x,y,w,h,radius,color
    tft_setTextColor(text_color);
    tft_setCursor(0, v_pos);
    tft_setTextSize(2);
    tft_writeString(print_buffer);
}

// Predefined colors definitions (from tft_master.h)
//#define	ILI9340_BLACK   0x0000
//#define	ILI9340_BLUE    0x001F
//#define	ILI9340_RED     0xF800
//#define	ILI9340_GREEN   0x07E0
//#define ILI9340_CYAN    0x07FF
//#define ILI9340_MAGENTA 0xF81F
//#define ILI9340_YELLOW  0xFFE0
//#define ILI9340_WHITE   0xFFFF

// === thread structures ============================================
// thread control structs
// note that UART input and output are threads
static struct pt pt_timer, pt_serial, pt_control;
// The following threads are necessary for UART control
static struct pt pt_input, pt_output, pt_DMA_output;



//static PT_THREAD (protothread_control(struct pt *pt))
//{
//    PT_BEGIN(pt);
//      while(1) {
//          if (modeFlag == 0){
//              
//          }
//          PT_YIELD(pt);
//      } // END WHILE(1)
//  PT_END(pt);
//} // timer thread

// system 1 second interval tick
int sys_time_seconds;
volatile char cmd[2000];
// === Timer Thread =================================================
// update a 1 second tick counter

static PT_THREAD(protothread_timer(struct pt *pt)) {
    PT_BEGIN(pt);
    // set up LED to blink
    //mPORTASetBits(BIT_0 );	//Clear bits to ensure light is off.
    //mPORTASetPinsDigitalOut(BIT_0 );    //Set port as output
    while (1) {
        // yield time 1 second
        PT_YIELD_TIME_msec(1000);
        sys_time_seconds++;
        // toggle the LED on the big board
        //mPORTAToggleBits(BIT_0);
        // draw sys_time
        sprintf(buffer, "Time=%d", sys_time_seconds);
        printLine2(0, buffer, ILI9340_BLACK, ILI9340_YELLOW);
        //------------Debug
        /*
        sprintf(buffer,"Draw Counter=%d", drawCounter);
        printLine2(2, buffer, ILI9340_BLACK, ILI9340_YELLOW);
        sprintf(buffer,"msCounter=%d", msCounter);
        printLine2(3, buffer, ILI9340_BLACK, ILI9340_YELLOW);
        sprintf(buffer,"%c%c", cmd[0],cmd[1]);
        printLine2(6, buffer, ILI9340_BLACK, ILI9340_YELLOW);
        if (drawCounter>=0){
            sprintf(buffer,"%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c", PT_term_buffer[drawCounter*15],
                    PT_term_buffer[drawCounter*15+1],PT_term_buffer[drawCounter*15+2],
                    PT_term_buffer[drawCounter*15+3],PT_term_buffer[drawCounter*15+4],
                    PT_term_buffer[drawCounter*15+5],PT_term_buffer[drawCounter*15+6],
                    PT_term_buffer[drawCounter*15+7],PT_term_buffer[drawCounter*15+8],
                    PT_term_buffer[drawCounter*15+9],PT_term_buffer[drawCounter*15+10],
                    PT_term_buffer[drawCounter*15+11],PT_term_buffer[drawCounter*15+12],
                    PT_term_buffer[drawCounter*15+13],PT_term_buffer[drawCounter*15+14]);
            printLine2(4, buffer, ILI9340_BLACK, ILI9340_YELLOW);
        }
         */
        //sprintf(PT_send_buffer,"Hello world \n\r", sys_time_seconds);
        // by spawning a print thread
        //PT_SPAWN(pt, &pt_DMA_output, PT_DMA_PutSerialBuffer(&pt_DMA_output) );

        // NEVER exit while
    } // END WHILE(1)
    PT_END(pt);
} // timer thread


//Print out in binary

void binaryPrint(void) {
    unsigned char binaryBuffer[16];
    int loop, shift;
    for (loop = 15; loop >= 0; loop--) {
        if (loop > 7) {
            shift = loop - 8;
            if (!!(cmd[1] & (1 << shift))) binaryBuffer[loop] = 1;
            else binaryBuffer[loop] = 0;
        } else {
            shift = loop;
            if (!!(cmd[2] & (1 << shift))) binaryBuffer[loop] = 1;
            else binaryBuffer[loop] = 0;
        }
    }
    sprintf(PT_send_buffer, "%u%u%u%u %u%u%u%u %u%u%u%u %u%u%u%u\n\r", binaryBuffer[15],
            binaryBuffer[14], binaryBuffer[13], binaryBuffer[12], binaryBuffer[11],
            binaryBuffer[10], binaryBuffer[9], binaryBuffer[8], binaryBuffer[7],
            binaryBuffer[6], binaryBuffer[5], binaryBuffer[4], binaryBuffer[3],
            binaryBuffer[2], binaryBuffer[1], binaryBuffer[0]);
}
// === Direct Play Mode Helper Function=======================================
// "setEachValve" helper function to set a digital pin for a valve

void setEachValve(int num, int on) {
    if (on) {
        if (num == 0) mPORTASetBits(BIT_0);
        else if (num == 1) mPORTBSetBits(BIT_3);
        else if (num == 2) mPORTASetBits(BIT_2);
        else if (num == 3) mPORTASetBits(BIT_3);
        else if (num == 4) mPORTBSetBits(BIT_4);
        else if (num == 5) mPORTASetBits(BIT_4);
        else if (num == 6) mPORTBSetBits(BIT_5);
        else if (num == 7) mPORTBSetBits(BIT_15);
        else if (num == 8) mPORTBSetBits(BIT_13);
        else if (num == 9) mPORTBSetBits(BIT_9);
        else if (num == 10) mPORTBSetBits(BIT_8);
        else if (num == 11) mPORTBSetBits(BIT_7);
    } else {
        if (num == 0) mPORTAClearBits(BIT_0);
        else if (num == 1) mPORTBClearBits(BIT_3);
        else if (num == 2) mPORTAClearBits(BIT_2);
        else if (num == 3) mPORTAClearBits(BIT_3);
        else if (num == 4) mPORTBClearBits(BIT_4);
        else if (num == 5) mPORTAClearBits(BIT_4);
        else if (num == 6) mPORTBClearBits(BIT_5);
        else if (num == 7) mPORTBClearBits(BIT_15);
        else if (num == 8) mPORTBClearBits(BIT_13);
        else if (num == 9) mPORTBClearBits(BIT_9);
        else if (num == 10) mPORTBClearBits(BIT_8);
        else if (num == 11) mPORTBClearBits(BIT_7);
    }
}

// "setValves" helper function to call "setEachValve" to set all digital pins for all 15 valves

void setValves(void) {
    int counter;
    for (counter = 0; counter < 12; counter++) { //Counter from 0 to 11
        if (cmd[counter] == '1') { //50 or 0011 0001
            setEachValve(counter, 1);
            //sprintf(PT_send_buffer, "%d \n\r", counter);
            //printLine2(10, PT_send_buffer, ILI9340_YELLOW , ILI9340_BLACK);
        } else { //(cmd[counter] == '0') //49 or 0011 0000
            setEachValve(counter, 0);
            //sprintf(PT_send_buffer, "%d \n\r", counter);
            //printLine2(11, PT_send_buffer, ILI9340_YELLOW , ILI9340_BLACK);
        }
    }
}

// "setValvesProgramMode" helper function to call "setEachValve" to set all digital pins for all 12 valves iteratively

void setValvesProgramMode(int baseValue) {
    int counter;
    int counter2 = 0;
    for (counter = baseValue; counter < (baseValue + 12); counter++) { //Counter from 0 to 11
        if (PT_term_buffer[counter] == '1') { //50 or 0011 0001
            setEachValve(counter2, 1);
            //sprintf(PT_send_buffer, "%d \n\r", counter);
            //printLine2(10, PT_send_buffer, ILI9340_YELLOW , ILI9340_BLACK);
        } else { //(cmd[counter] == '0') //49 or 0011 0000
            setEachValve(counter2, 0);
            //sprintf(PT_send_buffer, "%d \n\r", counter);
            //printLine2(11, PT_send_buffer, ILI9340_YELLOW , ILI9340_BLACK);
        }
        counter2++;
    }
}

// "setValvesShow" helper function calls "setEachValve" to set all digital pins for all 12 valves

void setValvesShow(void) {
    int counter;
    for (counter = 0; counter < 12; counter++) { //Counter from 0 to 11
        if (showBuffer[counter] == 1) { //Stop water fall
            setEachValve(counter, 1);
            //sprintf(PT_send_buffer, "%d \n\r", counter);
            //printLine2(10, PT_send_buffer, ILI9340_YELLOW , ILI9340_BLACK);
        } else { //(showBuffer[counter] == 0) //49 or 0011 0000
            setEachValve(counter, 0); //Enable water fall
        }
    }
}

void stopShowBuffer(void) {
    int loop;
    for (loop = 0; loop < 12; loop++) {
        showBuffer[loop] = 1;
    }
}

void enjoyModeInitialize(void){
    msCounter = 0;
    msCounter2 = 0;
    showCounter = 0;
    showNum = 1;
    doubleCounter = 0;
    show8Counter = 0;
    show8Counter2 = 0;
    show9Counter = 0;
    show9up = 1;
    show11Counter = 0;
}

//=== Serial terminal thread =================================================
static PT_THREAD(protothread_serial(struct pt *pt)) {
    PT_BEGIN(pt);
    while (1) {
        //Spawn a thread to read data from serial ports
        //Return data received to PT_term_buffer when '/r' is received
        PT_SPAWN(pt, &pt_input, PT_GetSerialBuffer(&pt_input));
        
        //Scan for the number of element in the receiving buffer
        for (bufferCounter = 0; bufferCounter < 350; bufferCounter++) {
            if (PT_term_buffer[bufferCounter] == '2') {
                break;
            }
        }
        
        //Decode and enter different mode based on the size of PT_term_buffer
        if (bufferCounter == 12) {         //Play Mode
            modeFlag = 0;
            setValves(); //Set valves in play mode
        } else if (bufferCounter == 240) { //Program Mode
            modeFlag = 1;
            drawCounter = 19;
            msCounter = 0;
        } else if (bufferCounter == 20) {  //Enjoy Mode
            modeFlag = 2;
            enjoyModeInitialize(); //Initialize variables in enjoy mode
        }
    }
    PT_END(pt);
}

// === Main  ======================================================

void main(void) {
    ANSELA = 0;
    ANSELB = 0;
    /*
  1	RA0	
  2	RA1	RX on big board
  3	RB0	LED1         --
  4	RB1	LED2
  5	RB2	LED3
  6	RB3	LED4
  7	RA2	LED5 
  8	RA3	LED6
  9	RB4	LED7
  10	RA4	LED8
  11	RB5	
  12	RB15	
  13	RB14	
  14	RB13	
  15	RB12	
  16	RB11	
  17	RB10	TX on big board
  18	RB9	
  19	RB8	
  20	RB7	
  21	RB6	
     */
    //Set up digital pins as output and initialize them to LOW
    mPORTBSetPinsDigitalOut(BIT_3 | BIT_4 | BIT_5 | BIT_13 | BIT_15 | BIT_9 | BIT_8 | BIT_7);
    mPORTBClearBits(BIT_3 | BIT_4 | BIT_5 | BIT_13 | BIT_15 | BIT_9 | BIT_8 | BIT_7);
    mPORTASetPinsDigitalOut(BIT_2 | BIT_3 | BIT_4 | BIT_0);
    mPORTAClearBits(BIT_2 | BIT_3 | BIT_4 | BIT_0);
    //The average reaction time for humans is 0.25 seconds to a visual stimulus
    //0.17 for an audio stimulus, and 0.15 seconds for a touch stimulus.

    //Set up timer interrupt
    // Set up timer2 on,  interrupts, internal clock, prescalar 1, toggle rate
    // at 30 MHz PB clock 60 counts is two microsec
    // 400 is 100 ksamples/sec
    // 2000 is 20 ksamp/sec
    // 0.7m == 0.38 sec of free fall
    // 0.5m == 0.32 sec of free fall
    // 40 MHz or 40,000,000Hz = 1 sec
    // 0.01 sec or 10ms = 400,000
    // 32 * 400,000 / 20 resolution = 6400000
    //40e6 / (20 resolution / 0.32 fall time)
    //TMR is only 16 bit or 65536 
    OpenTimer2(T2_ON | T2_SOURCE_INT | T2_PS_1_1, 40000); //1ms 
    //OpenTimer2(T2_ON | T2_SOURCE_INT | T2_PS_1_1, 2000000);

    // set up the timer interrupt with a priority of 2
    ConfigIntTimer2(T2_INT_ON | T2_INT_PRIOR_2);
    mT2ClearIntFlag(); // and clear the interrupt flag    


    // === config threads ==========
    PT_setup();

    // === setup system wide interrupts  ========
    INTEnableSystemMultiVectoredInt();

    // init the threads
    PT_INIT(&pt_timer);
    PT_INIT(&pt_serial);

    // init the display
    // NOTE that this init assumes SPI channel 1 connections
    tft_init_hw();
    tft_begin();
    tft_fillScreen(ILI9340_BLACK);
    //240x320 vertical display
    tft_setRotation(1); // Use tft_setRotation(1) for 320x240

    // seed random color
    srand(1);

    // round-robin scheduler for threads
    while (1) {
        PT_SCHEDULE(protothread_timer(&pt_timer));
        PT_SCHEDULE(protothread_serial(&pt_serial));
    }
} // main

// === end  ======================================================

